/**
 * Optiboot bootloader for Arduino
 *
 * http://optiboot.googlecode.com
 * This version is meant for tinyAVR-0, tinyAVR-1 and tinyAVR-2 devices, and
 * has USB functionality. This does mean that you need at least 4KiB of flash
 * to fit the bootloader (which is 2560 bytes).
 */

/* Version that this USB bootloader is based on. */
#define OPTIBOOT_MAJVER      9
#define OPTIBOOT_MINVER      1

/**
 * USB CDC ACM bulk in and out transfer size, `in' is the input of the driver,
 * so it is our output transfer size, and `out' is the output of the driver, and
 * therefore our input, which is a little bit counter intuitive.
 */
#define HW_CDC_BULK_OUT_SIZE 1
#define HW_CDC_BULK_IN_SIZE  8

/* Reporting the serial state (carrier detect) is not really required (by linux at least). */
#define REPORT_SERIAL_STATE  0

/**
 * OPTIBOOT_CUSTOMVER should be defined (by the makefile) for custom edits
 * of optiboot.  That way you don't wind up with very different code that
 * matches the version number of a "released" optiboot.
 */
#define OPTIBOOT_CUSTOMVER   0x10

unsigned const int __attribute__((section(".version"))) __attribute__((used))
optiboot_version = 256 * (OPTIBOOT_MAJVER + OPTIBOOT_CUSTOMVER) + OPTIBOOT_MINVER;

#include <inttypes.h>
#include <stdbool.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/atomic.h>
#include <util/delay.h>

#if (!defined(__AVR_XMEGA__)) || (__AVR_ARCH__ != 103)
#   error CPU not supported by this version of Optiboot.
#endif

/**
 * stk500.h contains the constant definitions for the stk500v1 comm protocol.
 */
#include "stk500.h"

/**
 * Watchdog timeout translations from human readable to config values.
 */
#ifndef WDTTIME
#   define WDTPERIOD WDT_PERIOD_1KCLK_gc  // 1 second
#elif WDTTIME == 1
#   define WDTPERIOD WDT_PERIOD_1KCLK_gc  // 1 second
#elif WDTTIME == 2
#   define WDTPERIOD WDT_PERIOD_2KCLK_gc  // 2 seconds
#elif WDTTIME == 4
#   define WDTPERIOD WDT_PERIOD_4KCLK_gc  // 4 seconds
#elif WDTTIME == 8
#   define WDTPERIOD WDT_PERIOD_8KCLK_gc  // 8 seconds
#else
#endif

#include "usbdrv.h"

/**
 * RAMSTART should be self-explanatory. It's bigger on parts with a lot of peripheral registers.
 * NOTE that RAMSTART (for optiboot) need not be exactly at the start of RAM.
 */
#if !defined(RAMSTART)  // Newer versions of gcc avr-libc define RAMSTART.
#   error RAMSTART not defined.
#endif


#if MAPPED_PROGMEM_PAGE_SIZE > 255
typedef uint16_t pagelen_t;
#else
typedef uint8_t pagelen_t;
#endif

#include "usbdrv.h"

enum usb_request {
    SEND_ENCAPSULATED_COMMAND = 0,
    GET_ENCAPSULATED_RESPONSE,
    SET_COMM_FEATURE,
    GET_COMM_FEATURE,
    CLEAR_COMM_FEATURE,
    SET_LINE_CODING = 0x20,
    GET_LINE_CODING,
    SET_CONTROL_LINE_STATE,
    SEND_BREAK
};

enum stk_state {
    NEW_COMMAND,
    GET_PARAMETERS,
    VERIFY_SPACE,
};

FUSES = {
    .WDTCFG = PERIOD_OFF_gc | WINDOW_OFF_gc, // Disable Watchdog.
    .BODCFG = LVL_BODLEVEL2_gc | SAMPFREQ_1KHz_gc | ACTIVE_ENABLED_gc | SLEEP_DIS_gc, // 2.6V brown out detection.
    .OSCCFG = FREQSEL_16MHZ_gc, // Run at 16MHz instead of 20MHz.
#ifdef FUSE_TCD0CFG_DEFAULT
    .TCD0CFG = FUSE_TCD0CFG_DEFAULT,
#endif
#ifdef RSTPIN
    .SYSCFG0 =  CRCSRC_NOCRC_gc | RSTPINCFG_RST_gc | FUSE_EESAVE_bm,
#else
#   ifdef FUSE_RSTPINCFG_gm  // Group mask will be defined for triple-func pins.
    .SYSCFG0 =  CRCSRC_NOCRC_gc | RSTPINCFG_UPDI_gc | FUSE_EESAVE_bm,
#   else
    .SYSCFG0 =  CRCSRC_NOCRC_gc | FUSE_EESAVE_bm,
#   endif
#endif
    .SYSCFG1 = SUT_8MS_gc,
    .APPEND = FUSE_APPEND_DEFAULT, // Default (0x00) means app end will be at end of flash.
    .BOOTEND = 0x0a // Boot end (which is also app start) will be at 0x0a00.
};

static enum stk_state state = NEW_COMMAND;
static uint8_t command;
static uint16_t command_bytes_received = 0;

/**
 * Optiboot uses several "address" variables that are sometimes byte pointers,
 * sometimes word pointers. sometimes 16bit quantities, and sometimes built
 * up from 8bit input characters. avr-gcc is not great at optimizing the
 * assembly of larger words from bytes, but we can use the usual union to
 * do this manually. Expanding it a little, we can also get rid of casts.
 */
static union {
    uint8_t  *bptr;
    uint16_t *wptr;
    uint16_t  word;
    uint8_t   bytes[2];
} address = {0};
static pagelen_t length = 0;
static uint8_t which = 0;
static bool put_characters_later = false;

static bool send_empty_frame;
static bool watchdog_configured = false;
#if REPORT_SERIAL_STATE
static uint8_t intr3_status;    /* used to control interrupt endpoint transmissions */
#endif
static uint8_t mode_buffer[7];

static uint8_t put_buffer[HW_CDC_BULK_IN_SIZE];
static uint8_t put_buffer_index = 0;

const PROGMEM char usbDescriptorConfiguration[] = { /* USB CDC configuration descriptor. */
    9,          /* sizeof(usbDescrConfig): length of descriptor in bytes */
    USBDESCR_CONFIG,    /* descriptor type */
    67,
    0,          /* total length of data returned (including inlined descriptors) */
    2,          /* number of interfaces in this configuration */
    1,          /* index of this configuration */
    0,          /* configuration name string index */
#if USB_CFG_IS_SELF_POWERED
    (char)((1 << 7) | USBATTR_SELFPOWER),     /* attributes */
#else
    (char)(1 << 7),                           /* attributes */
#endif
    USB_CFG_MAX_BUS_POWER / 2,                /* max USB current in 2mA units */

    /* interface descriptor follows inline: */
    9,          /* sizeof(usbDescrInterface): length of descriptor in bytes */
    USBDESCR_INTERFACE, /* descriptor type */
    0,          /* index of this interface */
    0,          /* alternate setting for this interface */
    USB_CFG_HAVE_INTRIN_ENDPOINT,   /* endpoints excl 0: number of endpoint descriptors to follow */
    USB_CFG_INTERFACE_CLASS,
    USB_CFG_INTERFACE_SUBCLASS,
    USB_CFG_INTERFACE_PROTOCOL,
    0,          /* string index for interface */

    /* CDC Class-Specific descriptor */
    5,           /* sizeof(usbDescrCDC_HeaderFn): length of descriptor in bytes */
    0x24,        /* descriptor type */
    0,           /* header functional descriptor */
    0x10, 0x01,

    4,           /* sizeof(usbDescrCDC_AcmFn): length of descriptor in bytes    */
    0x24,        /* descriptor type */
    2,           /* abstract control management functional descriptor */
    0x02,        /* SET_LINE_CODING, GET_LINE_CODING, SET_CONTROL_LINE_STATE    */

    5,           /* sizeof(usbDescrCDC_UnionFn): length of descriptor in bytes  */
    0x24,        /* descriptor type */
    6,           /* union functional descriptor */
    0,           /* CDC_COMM_INTF_ID */
    1,           /* CDC_DATA_INTF_ID */

    5,           /* sizeof(usbDescrCDC_CallMgtFn): length of descriptor in bytes */
    0x24,        /* descriptor type */
    1,           /* call management functional descriptor */
    3,           /* allow management on data interface, handles call management by itself */
    1,           /* CDC_DATA_INTF_ID */

    /* Endpoint Descriptor */
    7,           /* sizeof(usbDescrEndpoint) */
    USBDESCR_ENDPOINT,  /* descriptor type = endpoint */
    (char)(0x80 | USB_CFG_EP3_NUMBER),        /* IN endpoint number 3 */
    0x03,        /* attrib: Interrupt endpoint */
    8, 0,        /* maximum packet size */
    (char)USB_CFG_INTR_POLL_INTERVAL,        /* in ms */

    /* Interface Descriptor  */
    9,           /* sizeof(usbDescrInterface): length of descriptor in bytes */
    USBDESCR_INTERFACE,           /* descriptor type */
    1,           /* index of this interface */
    0,           /* alternate setting for this interface */
    2,           /* endpoints excl 0: number of endpoint descriptors to follow */
    0x0A,        /* Data Interface Class Codes */
    0,
    0,           /* Data Interface Class Protocol Codes */
    0,           /* string index for interface */

    /* Endpoint Descriptor */
    7,           /* sizeof(usbDescrEndpoint) */
    USBDESCR_ENDPOINT,  /* descriptor type = endpoint */
    0x01,        /* OUT endpoint number 1 */
    0x02,        /* attrib: Bulk endpoint */
    HW_CDC_BULK_OUT_SIZE, 0,        /* maximum packet size */
    0,           /* in ms */

    /* Endpoint Descriptor */
    7,           /* sizeof(usbDescrEndpoint) */
    USBDESCR_ENDPOINT,  /* descriptor type = endpoint */
    (char)0x81,  /* IN endpoint number 1 */
    0x02,        /* attrib: Bulk endpoint */
    HW_CDC_BULK_IN_SIZE, 0,        /* maximum packet size */
    0,           /* in ms */
};

/* Everything that needs to run VERY early. */
void __attribute__((naked)) __attribute__((section(".init8"))) __attribute__((used)) pre_main(void) {
    // Allow convenient way of calling do_spm function - jump table,
    // so entry to this function will always be here, independent of compilation, features, etc.
    asm volatile (
        "   rjmp 1f\n"
#ifndef APP_NOSPM
        "   rjmp do_nvmctrl\n"
#else
        "   ret\n"   // If do_nvmctrl isn't required by the app, return without doing anything.
#endif
        "1:\n"
    );
}

static void __attribute__ ((noinline)) put_char_in_buffer(char ch) {
    put_buffer[put_buffer_index++] = (uint8_t)ch;
}

static bool __attribute__ ((noinline)) set_state_to_verify_space_if_n_bytes_received(uint8_t n) {
    if (++command_bytes_received >= n) {
        state = VERIFY_SPACE;
        return true;
    }
    return false;
}

#if USB_CFG_HAVE_MEASURE_FRAME_LENGTH
static void __attribute__ ((noinline)) usb_calibrate_osc() {
    // For 16.5MHz, the target frame length will be 2356 (we want to arrive at this point).
    // For 16.0MHz, the target frame length will be 2284 (internal oscillator at 16MHz starts at about this point).
    // For 12.8MHz, the target frame length will be 1827 (we want to arrive at this point).
    int16_t target_frame_len = (unsigned int)(1499. * (double)F_CPU / 10.5e6 + 0.5);
    uint8_t  tmp = CLKCTRL.OSC20MCALIBA; // Oscilator default calibration value.

    // Trim to 16.5MHz or 12.8MHz
    // https://www.silabs.com/community/interface/knowledge-base.entry.html/2004/03/15/usb_clock_tolerance-gVai
    // http://vusb.wikidot.com/examples
    uint8_t  sav = tmp;
    int16_t low = 32767; // Lowest saved value.
    for (;;) { // Normally solves in about 5 iterations.
        // We expect cur to be negative numbers until we overshoot.
        int16_t cur = usbMeasureFrameLength() - target_frame_len; 

        int16_t cur_abs = cur;
        if (cur_abs < 0) 
            cur_abs = -cur_abs; // Make a positive number, so we know how far away from zero it is.
        if (cur_abs < low) { // Current value is lower.
            low = cur_abs; // Got a new lower value, save it.
            sav = tmp; // Save the OSC20MCALIBA value.
        } else { // Things just got worse, curabs > low, so saved value before this was the best value.
            _PROTECTED_WRITE(CLKCTRL.OSC20MCALIBA, sav);
            return;
        }

        if (cur < 0) // Freq too high.
            tmp++;  // +1 will increase frequency by about 1.5%
        else        // Freq too low
            tmp--;  // -1 will increase frequency by about 1.5%
        // Apply change and check again.
        _PROTECTED_WRITE(CLKCTRL.OSC20MCALIBA, tmp);
    }
}
#endif

/**
 * Change the watchdog configuration.
 * Could be a new timeout, could be off...
 */
void __attribute__ ((noinline)) watchdog_config(uint8_t period) {
    while (WDT.STATUS & WDT_SYNCBUSY_bm); // Busy wait for sycnhronization is required!
    _PROTECTED_WRITE(WDT.CTRLA, period);
}


uchar usbFunctionDescriptor(usbRequest_t *rq) {
    if (rq->wValue.bytes[1] == USBDESCR_DEVICE) {
        usbMsgPtr = (usbMsgPtr_t)usbDescriptorDevice;
        return usbDescriptorDevice[0];
    } else {  /* Must be config descriptor. */
        usbMsgPtr = (usbMsgPtr_t)usbDescriptorConfiguration;
        return sizeof(usbDescriptorConfiguration);
    }
}

usbMsgLen_t usbFunctionSetup(uchar data[8]) {
    usbRequest_t *rq = (void *)data;

    if ((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS) { /* Class request type. */
        if ((enum usb_request)rq->bRequest == GET_LINE_CODING || (enum usb_request)rq->bRequest == SET_LINE_CODING) {
            return 0xff;
            /* GET_LINE_CODING -> usbFunctionRead() */
            /* SET_LINE_CODING -> usbFunctionWrite() */
        }
        if ((enum usb_request)rq->bRequest == SET_CONTROL_LINE_STATE) {
            /**
             * Report serial state (carrier detect). On several Unix platforms,
             * tty devices can only be opened when carrier detect is set.
             */
#if REPORT_SERIAL_STATE
            if (intr3_status == 0)
                intr3_status = 2;
#endif
        }

        /*  Prepare bulk-in endpoint to respond to early termination   */
        if ((rq->bmRequestType & USBRQ_DIR_MASK) == USBRQ_DIR_HOST_TO_DEVICE)
            send_empty_frame  = true;
    }

    return 0;
}

uchar usbFunctionRead(uchar *data, uchar len) {
    (void)len;
    /* GET_LINE_CODING */
    data[0] = mode_buffer[0];
    data[1] = mode_buffer[1];
    data[2] = mode_buffer[2];
    data[3] = mode_buffer[3];
    data[4] = mode_buffer[4];
    data[5] = mode_buffer[5];
    data[6] = mode_buffer[6];
    return 7;
}

uchar usbFunctionWrite(uchar *data, uchar len) {
    (void)len;
    /* SET_LINE_CODING */
    mode_buffer[0] = data[0];
    mode_buffer[1] = data[1];
    mode_buffer[2] = data[2];
    mode_buffer[3] = data[3];
    mode_buffer[4] = data[4];
    mode_buffer[5] = data[5];
    mode_buffer[6] = data[6];
    return 1;
}

void usbFunctionWriteOut(uchar *data, uchar len) {
    (void)len;
    if (!watchdog_configured) {
        watchdog_configured = true;
        // Set up watchdog to trigger after a bit (nominally:, 1s for autoreset, longer for manual)
        watchdog_config(WDTPERIOD);
    }
    // Reset watchdog.
    asm volatile ("wdr\n");
#if USB_CFG_HAVE_FLOWCONTROL
    usbDisableAllRequests();
#endif
#if HW_CDC_BULK_OUT_SIZE > 1
    for (uint8_t i = 0; i < len; i++)
#else
    uint8_t i = 0;
#endif
    {
        uint8_t ch = data[i];
        switch (state) {
        case NEW_COMMAND:
            command = ch;
            switch (command) {
            case STK_GET_PARAMETER: case STK_SET_DEVICE: case STK_SET_DEVICE_EXT: case STK_LOAD_ADDRESS:
            case STK_UNIVERSAL: case STK_PROG_PAGE: case STK_READ_PAGE:
                state = GET_PARAMETERS;
                break;
            case STK_LEAVE_PROGMODE: /* 'Q' */
                // Adaboot no-wait mod.
                watchdog_config(WDT_PERIOD_8CLK_gc);
                /* fallthrough */
            case STK_READ_SIGN:
            default: // This covers the response to commands like STK_ENTER_PROGMODE.
                state = VERIFY_SPACE;
            }
            command_bytes_received = 0;
            break;
        case GET_PARAMETERS:
            switch (command) {
            case STK_GET_PARAMETER:
                which = ch;
                state = VERIFY_SPACE;
                break;
            case STK_SET_DEVICE: // SET DEVICE is ignored.
                set_state_to_verify_space_if_n_bytes_received(20);
                break;
            case STK_SET_DEVICE_EXT: // SET DEVICE EXT is ignored.
                set_state_to_verify_space_if_n_bytes_received(5);
                break;
            case STK_LOAD_ADDRESS: // LOAD ADDRESS.
                address.bytes[set_state_to_verify_space_if_n_bytes_received(2) ? 1 : 0] = ch;
                // TODO: will there be mega-0 chips with >128k of RAM?
                /**
                 * UPDI chips apparently have byte-addressable FLASH?
                 * address.word *= 2; // Convert from word address to byte address.
                 */
                break;
            case STK_UNIVERSAL: // UNIVERSAL command is ignored.
                if (set_state_to_verify_space_if_n_bytes_received(4))
                    put_char_in_buffer(0x00);
                break;
            case STK_PROG_PAGE: case STK_READ_PAGE:
                ++command_bytes_received;
                switch (command_bytes_received) {
                /**
                 * We can never load flash with more than 1 page at a time, so we can save
                 * some code space on parts with smaller pagesize by using a smaller int.
                 */
#if MAPPED_PROGMEM_PAGE_SIZE > 255
                case 1:
                    length = (pagelen_t)ch << 8;
                    break;
                case 2:
                    length |= ch;
                    break;
#else
                case 1:
                    break;
                case 2:
                    length = ch;
                    break;
#endif
                case 3:
                    // PROGRAM PAGE - any kind of page!
                    if (ch == 'F') {
                        address.word += MAPPED_PROGMEM_START;
                        // WARNING: This is a workaround because AVRDUDE thinks that we are an ATtiny202 with
                        // a 512 byte bootloader while we are actually an ATtiny402 with a 2560 byte bootloader.
                        address.word += (0x0a00 - 0x0200);
                    } else {
                        address.word += MAPPED_EEPROM_START;
                    }
                    // TODO: user row?
                    if (command == STK_READ_PAGE)
                        state = VERIFY_SPACE;
                    break;
                default:
                    *(address.bptr++) = ch;
                    if (command_bytes_received - 3 >= length)
                        state = VERIFY_SPACE; // Read command terminator, start reply
                }
                break;
            }
            break;
        case VERIFY_SPACE:
            if (ch != CRC_EOP) {
                watchdog_config(WDT_PERIOD_8CLK_gc); // Shorten watchdog timeout...
                for (;;); // ...and busy-loop so that WDT causes a reset and app start.
            }
            put_char_in_buffer(STK_INSYNC);
            switch (command) {
            case STK_GET_PARAMETER:
                /**
                 * Send optiboot version as "SW version"
                 * NOTE that the references to memory are optimized away.
                 */
                switch (which) {
                case STK_SW_MINOR:
                    put_char_in_buffer(optiboot_version & 0xFF);
                    break;
                case STK_SW_MAJOR:
                    put_char_in_buffer(optiboot_version >> 8);
                    break;
                default:
                    /* GET PARAMETER returns a generic 0x03 reply for other parameters; enough to keep AVRDUDE happy. */
                    put_char_in_buffer(0x03);
                }
                break;
            case STK_PROG_PAGE: /* Write memory, length is big endian and is in bytes. */
                /* Actually Write the buffer to flash (and wait for it to finish). */
                _PROTECTED_WRITE_SPM(NVMCTRL.CTRLA, NVMCTRL_CMD_PAGEERASEWRITE_gc);
                // Wait for flash and EEPROM not busy, just in case.
                while (NVMCTRL.STATUS & (NVMCTRL_FBUSY_bm | NVMCTRL_EEBUSY_bm));
                break;
            case STK_READ_PAGE: /* Read memory block mode, length is big endian.  */
                put_characters_later = true;
                break;
            case STK_READ_SIGN: /* Get device signature bytes.  */
                // READ SIGN - return what AVRDUDE wants to hear.
                // WARNING: ATtiny402 workaround to let AVRDUDE think that we are an ATtiny202,
                // so that a bootloader size of 512 bytes seems correct, while we are actually using 2560 bytes.
                put_char_in_buffer((char)0x1E);
                put_char_in_buffer((char)0x91);
                put_char_in_buffer((char)0x23);
                break;
            }

            if (command != STK_READ_PAGE) {
                put_char_in_buffer(STK_OK);
                state = NEW_COMMAND;
            }
            break;
        }
    }
}

#if USB_CFG_HAVE_MEASURE_FRAME_LENGTH
void usbHadReset() {
    // usbMeasureFrameLength() counts CPU cycles, so disable interrupts.
    ATOMIC_BLOCK(ATOMIC_FORCEON) {
        usb_calibrate_osc();
    }
}
#endif

#ifndef APP_NOSPM
/**
 * Separate function for doing nvmctrl stuff.
 * It's needed for application to do manipulate flash, since only the
 * bootloader can write or erase flash, or write to the flash alias areas.
 * Note that this is significantly different in the details than the
 * do_spm() function provided on older AVRs.  Same "vector", though.

 * How it works:
 * - if the "command" is legal, write it to NVMCTRL.CTRLA
 * - if the command is not legal, store data to *address
 * - wait for NVM to complete

 * For example, to write a flash page:
 * Copy each byte with:
 * `do_nvmctrl(flash_offset + MAPPED_PROGMEM_START, 0xFF, *input_ptr);'
 * Erase and write page with
 * `do_nvmctrl(0, NVMCTRL_CMD_PAGEERASEWRITE_gc, 0);'
 */

static void __attribute__((used)) do_nvmctrl(uint16_t address, uint8_t command, uint8_t data) {
    if (command <= NVMCTRL_CMD_gm) {
        _PROTECTED_WRITE_SPM(NVMCTRL.CTRLA, command);
        // Wait for flash and EEPROM not busy, just in case.
        while (NVMCTRL.STATUS & (NVMCTRL_FBUSY_bm | NVMCTRL_EEBUSY_bm));
    } else {
        *(uint8_t *)address = data;
    }
}
#endif

/* Main program starts here. */
int main(void) {
    /**
     * This is the first code to run after the libc start files.
     * 
     * Optiboot relies on the peripherals not being configured in an adverse way.
     * Everything is written with the assumption that we start out after a reset.
     * So what do we do when we can see that it isn't arriving from a reset, because RSTFR is 0?
     */
    uint8_t ch = RSTCTRL.RSTFR; // We get the reset cause...
    if (ch == 0) {
        // ... and if there is no reset cause that would mean that this isn't the first code to run,
        // so we do a software reset.
        _PROTECTED_WRITE(RSTCTRL.SWRR, RSTCTRL_SWRE_bm);
    }
  
    /**
     * Setup main clk freq to 16MHz, we trim to 16.5MHz later. Programming fuses will set BOD, and 16MHz or 20MHz.
     * Use 0x02 (OSCCFG) value must be 0x02, for 16MHz fuse must be 0x01.
     */
    
    // No prescaler = 0x00, prescaler enable = CLKCTRL_PEN_bm (default prescaler is div6).
    _PROTECTED_WRITE(CLKCTRL.MCLKCTRLB, (CLKCTRL.MCLKCTRLB & ~CLKCTRL_PEN_bm));
    // Wait for OSC20MS to become stable.
    while (!(CLKCTRL.MCLKSTATUS & CLKCTRL_OSC20MS_bm));

    PORTA.DIR &= ~PIN3_bm; // PORTA_PIN3 as input for the button.
    PORTA.PIN3CTRL |= PORT_PULLUPEN_bm;
    _delay_us(10); // Wait for the button pin to stabilise.
    
#if !defined(ENTRYCOND_REQUIRE)
// Default bootloader entry condition is power on reset, external reset, software reset or UDPI reset.
#   define ENTRYCOND_REQUIRE (RSTCTRL_PORF_bm | RSTCTRL_EXTRF_bm | RSTCTRL_SWRF_bm | RSTCTRL_UPDIRF_bm)
#endif
    
    // Go to app if we had a watchdog reset, if the bootloader entry condition was not met or
    // if the button is not pressed.
    if ((ch & RSTCTRL_WDRF_bm) || !(ch & ENTRYCOND_REQUIRE) || (PORTA.IN & PIN3_bm)) {
        /**
         * Start the app. Don't bother trying to stuff it in r2, which requires heroic effort to fish out,
         * we'll put it in GPIOR0 where it won't get stomped on.
         */
        // asm volatile ("mov r2, %0\n" :: "r" (ch));
        RSTCTRL.RSTFR = ch; // Clear the reset causes before jumping to app...
        GPIOR0 = ch; // ...but stash the reset cause in GPIOR0 for use by app.
        watchdog_config(WDT_PERIOD_OFF_gc);
        asm volatile( "jmp 0x0a00\n\t");
    }

    // Relocate interrupt table to the start of the bootloader instead of start of app.
    _PROTECTED_WRITE(CPUINT.CTRLA, (CPUINT.CTRLA | CPUINT_IVSEL_bm));

    // Trim to 16.5MHz
    // https://www.silabs.com/community/interface/knowledge-base.entry.html/2004/03/15/usb_clock_tolerance-gVai
    // http://vusb.wikidot.com/examples
    // Test calibration trimming, to run at 16.5MHz... This test worked, we are running at 16.448Mhz, 
    // which is 0.316% of target freq. So we should be able to use this method for V-USB driver.
    // +2 will increase frequency by about 3.6%, and +1 by about 1.8%.
    _PROTECTED_WRITE(CLKCTRL.OSC20MCALIBA, CLKCTRL.OSC20MCALIBA + 1);
    // Wait for OSC20MS to become stable.
    while (!(CLKCTRL.MCLKSTATUS & CLKCTRL_OSC20MS_bm));

    // Disable interrupts.
    cli();
    
    usbInit();
    usbDeviceDisconnect(); // Enforce re-enumeration, do this while interrupts are disabled!
    _delay_ms(250);
    usbDeviceConnect();
#if REPORT_SERIAL_STATE
    intr3_status = 0;
#endif
    send_empty_frame  = false;

    // Re-enable interrupts.
    sei();

    /* Forever loop: exits by causing WDT reset. */
    for (;;) {
        usbPoll();
        bool flush = !put_characters_later;
        if (put_characters_later && put_buffer_index < HW_CDC_BULK_IN_SIZE) {
            put_char_in_buffer(*(address.bptr++));
            length--;
            if (length == 0) {
                put_characters_later = false;
                if (state == VERIFY_SPACE) {
                    put_char_in_buffer(STK_OK);
                    state = NEW_COMMAND;
                }
            }
        }

        if (usbInterruptIsReady()) {
            if (send_empty_frame) {
                usbSetInterrupt(put_buffer, 0);
                send_empty_frame = false;
                put_buffer_index = 0;
#if USB_CFG_HAVE_FLOWCONTROL
                if (usbAllRequestsAreDisabled())
                    usbEnableAllRequests();
#endif
            } else if (put_buffer_index >= HW_CDC_BULK_IN_SIZE || (flush && put_buffer_index)) {
                usbSetInterrupt(put_buffer, put_buffer_index);
                send_empty_frame = true;
            } else {
#if USB_CFG_HAVE_FLOWCONTROL
                if (usbAllRequestsAreDisabled())
                    usbEnableAllRequests();
#endif
            }
        }

#if REPORT_SERIAL_STATE
        /* We need to report rx and tx carrier after open attempt. */
        if (intr3_status != 0 && usbInterruptIsReady3()) {
            static uint8_t serialStateNotification[] = { 0xa1, 0x20, 0, 0, 0, 0, 2, 0, 3, 0 };

            if (intr3_status == 2) {
                usbSetInterrupt3(serialStateNotification, HW_CDC_BULK_IN_SIZE);
            } else {
                usbSetInterrupt3(serialStateNotification + HW_CDC_BULK_IN_SIZE, 
                                 sizeof(serialStateNotification) - HW_CDC_BULK_IN_SIZE);
            }
            intr3_status--;
        }
#endif
    }
}
