#include <stdbool.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <util/delay.h>

#include "strandtest_v4.h"
#include "neopixel.h"

#ifdef __AVR_ATtiny202__
#define FLASH_OFFSET_ERROR (0x0a00 - 0x0200)
// Don't use the maps to save space.
#define PIXELMAP1(x) (x)
#define PIXELMAP2(x) (x)
#define PIXELMAP_SPIRAL(x) ((uint8_t *)pixelmap_spiral + FLASH_OFFSET_ERROR)[x]
#define PIXELMAP_RANDOM(x) (x)

#define COLORS(x) ((uint32_t *)((uint8_t *)colors + FLASH_OFFSET_ERROR))[x]
#define FRMAP(x) ((uint8_t *)frmap + FLASH_OFFSET_ERROR)[x]
#else
#define PIXELMAP1(x) pixelmap1[x]
#define PIXELMAP2(x) pixelmap2[x]
#define PIXELMAP_SPIRAL(x) pixelmap_spiral[x]
#define PIXELMAP_RANDOM(x) pixelmap_random[x]

#define COLORS(x) colors[x]
#define FRMAP(x) frmap[x]
#endif

const uint8_t pixelmap1[NEO_NUM_PIXELS + 2] = {
    LED2, LED16, LED50, LED20, LED38, LED33,
    LED4,
    LED18, LED22, LED35, LED40, LED52, LED5,
    LED10, LED48, LED44, LED30, LED26, LED14,
    LED9,
    LED45, LED41, LED27, LED23, LED11, LED6,
    LED1, LED15, LED19, LED32, LED37, LED49,
    LED4,
    LED47, LED43, LED8, LED29, LED13, LED25,
    LED21, LED17, LED34, LED3, LED39, LED51,
    LED9,
    LED28, LED42, LED24, LED46, LED12, LED7,
};

const uint8_t pixelmap2[NEO_NUM_PIXELS + 2] = {
    LED2, LED7, LED16, LED12, LED50, LED46, LED20, LED24, LED38, LED42, LED33, LED28,
    LED4, LED9,
    LED18, LED14, LED22, LED26, LED35, LED30, LED40, LED44, LED52, LED48, LED5, LED10,
    LED1, LED6, LED15, LED11, LED19, LED23, LED32, LED27, LED37, LED41, LED49, LED45,
    LED9, LED4,
    LED47, LED51, LED43, LED39, LED8, LED3, LED29, LED34, LED13, LED17, LED25, LED21,
};

const uint8_t pixelmap_spiral[NEO_NUM_PIXELS] = {
    LED9, LED4, LED33, LED18, LED14, LED45, LED49, LED47, LED51, LED28, LED38, LED22, LED26, LED41, LED37, LED43, LED39,
    LED42, LED20, LED35, LED30, LED27, LED32, LED8, LED3, LED24, LED50, LED40, LED44, LED23, LED19, LED29, LED34, LED46,
    LED16, LED52, LED48, LED11, LED15, LED13, LED17, LED12, LED2, LED5, LED10, LED6, LED1, LED25, LED21, LED7
};

const uint8_t pixelmap_random[NEO_NUM_PIXELS] = {
    10,
    1,
    24,
    16,
    30,
    39,
    40,
    37,
    20,
    12,
    32,
    46,
    47,
    11,
    4,
    44,
    7,
    28,
    0,
    8,
    2,
    29,
    43,
    3,
    22,
    42,
    45,
    5,
    18,
    27,
    19,
    49,
    38,
    36,
    9,
    26,
    35,
    6,
    25,
    33,
    23,
    13,
    17,
    21,
    31,
    34,
    48,
    41,
    14,
    15,
};

const uint32_t colors[] = {
    0xFF0000, //Red
    0x00FF00, //Green
    0x0000FF, //Blue
    0x7F003F, //Purple
    0x7F3F00, //Orange
    0x007F1F, //Turquoise
    0x7F7F7F, //White
    0xFF500E  //Warm white
};

const uint8_t frmap[] = {
      1,   2,   3,   4,   5,   6,   7,   8, 
      9,  11,  12,  14,  15,  17,  18,  20, 
     21,  23,  25,  27,  29,  31,  33,  35, 
     37,  40,  43,  46,  49,  52,  55,  58, 
     61,  65,  69,  73,  77,  81,  85,  89, 
     93,  98, 103, 108, 113, 118, 123, 128, 
    134, 140, 146, 152, 158, 164, 170, 176, 
    182, 190, 198, 206, 214, 222, 230, 255
};

// IMPORTANT: To reduce NeoPixel burnout risk, add 1000 uF capacitor across
// pixel power leads, add 300 - 500 Ohm resistor on first pixel's data input
// and minimize distance between Arduino and first pixel.  Avoid connecting
// on a live circuit...if you must, connect GND first.

static volatile bool next_pattern = true, button_pressed_at_least_once = false;

static bool restoring_from_eeprom = false;
static uint8_t effect_num_on_boot;

ISR(PORTA_PORT_vect) {
    if (!(PORTA.INTFLAGS & PIN3_bm))
        return;
    if (!(PORTA.IN & PIN3_bm)) {
        next_pattern = button_pressed_at_least_once;
        button_pressed_at_least_once = true;
    }
    PORTA.INTFLAGS &= PIN3_bm;
}

static void set_next_pattern_false(int effect_num) {
    if (button_pressed_at_least_once) {
        // Turn off all leds first
        neopixel_clear();
        neopixel_show();
#ifndef __AVR_ATtiny202__
        if (effect_num_on_boot > 0) {
            restoring_from_eeprom = true;
            if (effect_num == effect_num_on_boot) {
                restoring_from_eeprom = false;
                next_pattern = false;
                effect_num_on_boot = 0;
            }
        } else
#endif
        {
            _delay_ms(200); // Simple button debounce
            next_pattern = false;
#ifndef __AVR_ATtiny202__
            eeprom_write_byte(0, effect_num);
#endif
        }
    }
}

// Input a value 0 to 255 to get a color value.
// The colours are a transition r - g - b - back to r.
static uint32_t wheel(uint8_t wheel_pos) {
    wheel_pos = 255 - wheel_pos;
    if (wheel_pos < 85) {
        return neopixel_color_rgb(255 - wheel_pos * 3, 0, wheel_pos * 3);
    }
    if (wheel_pos < 170) {
        wheel_pos -= 85;
        return neopixel_color_rgb(0, wheel_pos * 3, 255 - wheel_pos * 3);
    }
    wheel_pos -= 170;
    return neopixel_color_rgb(wheel_pos * 3, 255 - wheel_pos * 3, 0);
}

// Effect functions:
// Fill the dots one after the other with a color
static void color_wipe(uint32_t c, uint8_t wait) {
#ifdef __AVR_ATtiny202__
    uint8_t num_pixels = neopixel_get_num_pixels();
#else
    uint8_t num_pixels = neopixel_get_num_pixels() + 2;
#endif
    for (uint8_t i = 0; i < num_pixels; i++) {
        neopixel_set_pixel_color_c(PIXELMAP1(i), c);
        if (next_pattern && button_pressed_at_least_once)
            return;
        neopixel_show();
        _delay_ms(wait);
    }
}

// Fill the dots one after the other with a color
static void color_set(uint32_t c) {
    for (uint8_t i = 0; i < neopixel_get_num_pixels(); i++) {
        neopixel_set_pixel_color_c(i, c);
        if (next_pattern && button_pressed_at_least_once)
            return;
    }
    if (!next_pattern || !button_pressed_at_least_once)
        neopixel_show();
    while (!next_pattern);
}

static uint8_t __attribute__ ((noinline)) multiply_fixed_r(uint32_t c, uint8_t fr) {
    return (((c & 0xFF0000) >> 8) * FRMAP(fr)) >> 16;
}

static uint8_t __attribute__ ((noinline)) multiply_fixed_g(uint32_t c, uint8_t fr) {
    return ((c & 0x00FF00) * FRMAP(fr)) >> 16;
}

static uint8_t __attribute__ ((noinline)) multiply_fixed_b(uint32_t c, uint8_t fr) {
    return ((c & 0x0000FF) * FRMAP(fr)) >> 8;
}

// Fill the dots one after the other with a color and also turn them off again one after the other
static void color_wipe_on_off(uint32_t c, uint8_t wait) {
    bool first_time = true;
    do {
#ifdef __AVR_ATtiny202__
        uint8_t num_pixels = neopixel_get_num_pixels();
#else
        uint8_t num_pixels = neopixel_get_num_pixels() + 2;
#endif
        for (uint8_t i = 0; i < num_pixels; i++) {
            for (uint8_t j = 0; j < 64; j++) {
                uint8_t fr = ((j & 7) << 3) | (j >> 3);
                uint8_t fr_inv = 63 - fr;
                neopixel_set_pixel_color_rgb(PIXELMAP2((i + (7 - (j & 7))) % num_pixels),
                                             multiply_fixed_r(c, fr), multiply_fixed_g(c, fr), multiply_fixed_b(c, fr));
                                             
                uint8_t end_pos = i + (7 - (j & 7)) + 34;
                if (!first_time || end_pos >= num_pixels)
                    neopixel_set_pixel_color_rgb(PIXELMAP2((end_pos + num_pixels) % num_pixels),
                                                 multiply_fixed_r(c, fr_inv), multiply_fixed_g(c, fr_inv), multiply_fixed_b(c, fr_inv));
                if ((j & 7) == 7) {
                    neopixel_show();
                    _delay_us((((uint32_t)wait * 1000) >> 3) - 50);
                }
            }
            if (i >= 17)
                first_time = false;
            if (next_pattern && button_pressed_at_least_once)
                return;
        }
    } while (!next_pattern);
}

// Fill the dots one after the other with a color and also turn them off again one after the other
static void color_wipe_on_off_spiral(uint32_t c, uint8_t wait) {
    bool first_time = true;
    do {
        for (uint8_t i = 0; i < neopixel_get_num_pixels(); i++) {
            for (uint8_t j = 0; j < 64; j++) {
                uint8_t fr = ((j & 7) << 3) | (j >> 3);
                uint8_t fr_inv = 63 - fr;
                neopixel_set_pixel_color_rgb(PIXELMAP_SPIRAL((i + (7 - (j & 7))) % (neopixel_get_num_pixels())),
                                             multiply_fixed_r(c, fr), multiply_fixed_g(c, fr), multiply_fixed_b(c, fr));

                uint8_t end_pos = i + (7 - (j & 7)) + 32;
                if (!first_time || end_pos >= neopixel_get_num_pixels())
                    neopixel_set_pixel_color_rgb(PIXELMAP_SPIRAL((end_pos + neopixel_get_num_pixels()) % neopixel_get_num_pixels()),
                                                 multiply_fixed_r(c, fr_inv), multiply_fixed_g(c, fr_inv), multiply_fixed_b(c, fr_inv));
                if ((j & 7) == 7) {
                    neopixel_show();
                    _delay_us((((uint32_t)wait * 1000) >> 3) - 50);
                }
            }
            if (i >= 17)
                first_time = false;
            if (next_pattern && button_pressed_at_least_once)
                return;
        }
    } while (!next_pattern);
}

static void rainbow(uint8_t wait) {
    do {
        for (uint16_t j = 0; j < 256; j++) {
#ifdef __AVR_ATtiny202__
            uint8_t num_pixels = neopixel_get_num_pixels();
#else
            uint8_t num_pixels = neopixel_get_num_pixels() + 2;
#endif
            for (uint8_t i = 0; i < num_pixels; i++)
                neopixel_set_pixel_color_c(PIXELMAP2(i), wheel((i + j) & 255));
            neopixel_show();
            _delay_ms(wait);
            if (next_pattern && button_pressed_at_least_once)
                return;
        }
    } while (!next_pattern);
}

// Slightly different, this makes the rainbow equally distributed throughout
static void rainbow_cycle(uint8_t wait) {
    do {
        for (uint16_t j = 0; j < 256 * 5; j++) { // 5 cycles of all colors on wheel
#ifdef __AVR_ATtiny202__
            uint8_t num_pixels = neopixel_get_num_pixels();
#else
            uint8_t num_pixels = neopixel_get_num_pixels() + 2;
#endif
            for (uint8_t i = 0; i < num_pixels; i++)
                neopixel_set_pixel_color_c(PIXELMAP2(i), wheel(((i * 256 / num_pixels) + j) & 255));
            neopixel_show();
            _delay_ms(wait);
            if (next_pattern && button_pressed_at_least_once)
                return;
        }
    } while (!next_pattern);
}

// Slightly different, this makes the rainbow equally distributed throughout
static void rainbow_random_cycle(uint8_t wait) {
    do {
        for (uint16_t j = 0; j < 256 * 5; j++) { // 5 cycles of all colors on wheel
            for (uint8_t i = 0; i < neopixel_get_num_pixels(); i++)
                neopixel_set_pixel_color_c(PIXELMAP_RANDOM(i),
                                           wheel(((i * 256 / neopixel_get_num_pixels()) + j) & 255));
            neopixel_show();
            _delay_ms(wait);
            if (next_pattern && button_pressed_at_least_once)
                return;
        }
    } while (!next_pattern);
}

//Theatre-style crawling lights.
static void theatre_chase(uint32_t c, uint8_t wait) {
    do {
        for (uint8_t j = 0; j < 10; j++) { //do 10 cycles of chasing
            for (uint8_t q = 0; q < 3; q++) {
#ifdef __AVR_ATtiny202__
                uint8_t num_pixels = neopixel_get_num_pixels();
#else
                uint8_t num_pixels = neopixel_get_num_pixels() + 2;
#endif
                for (uint8_t i = 0; i < num_pixels; i = i + 3)
                    neopixel_set_pixel_color_c(PIXELMAP2(i + q), c);  //turn every third pixel on
                neopixel_show();

                _delay_ms(wait);

                for (uint8_t i = 0; i < num_pixels; i = i + 3)
                    neopixel_set_pixel_color_c(PIXELMAP2(i + q), 0);      //turn every third pixel off
                if (next_pattern && button_pressed_at_least_once)
                    return;
            }
        }
    } while (!next_pattern);
}

// Theatre-style crawling lights with rainbow effect
static void theatre_chase_rainbow(uint8_t wait) {
    do {
        for (uint16_t j = 0; j < 256; j++) {   // cycle all 256 colors in the wheel
            for (uint8_t q = 0; q < 3; q++) {
#ifdef __AVR_ATtiny202__
                uint8_t num_pixels = neopixel_get_num_pixels();
#else
                uint8_t num_pixels = neopixel_get_num_pixels() + 2;
#endif
                for (uint8_t i = 0; i < num_pixels; i = i + 3)
                    neopixel_set_pixel_color_c(PIXELMAP2(i + q), wheel((i + j) % 255)); //turn every third pixel on
                neopixel_show();

                _delay_ms(wait);

                for (uint8_t i = 0; i < num_pixels; i = i + 3)
                    neopixel_set_pixel_color_c(PIXELMAP2(i + q), 0);      //turn every third pixel off
                if (next_pattern && button_pressed_at_least_once)
                    return;
            }
        }
    } while (!next_pattern);
}

int main(void) {
    PORTA.DIR |= PIN6_bm; // PORTA_PIN6 as output for the neopixels.
    PORTA.DIR &= ~PIN3_bm; // PORTA_PIN3 as input for the button.
    PORTA.PIN3CTRL |= PORT_PULLUPEN_bm | PORT_ISC_FALLING_gc;

    neopixel_clear();
    neopixel_show(); // Initialize all pixels to 'off'
    neopixel_set_brightness(200);
#ifndef __AVR_ATtiny202__
    /* If the EEPROM was never written before, initialize to effect 1. */
    if (eeprom_read_byte(0) > NUM_EFFECTS)
        eeprom_write_byte(0, 1);
    effect_num_on_boot = eeprom_read_byte(0);
#endif
    while (1) {
        // Some example procedures showing how to display to the pixels:
        do {
            if (!restoring_from_eeprom) {
                color_wipe(neopixel_color_rgb(255, 0, 0), 100); //Red
                color_wipe(neopixel_color_rgb(127, 63, 0), 100); //Orange
                color_wipe(neopixel_color_rgb(0, 255, 0), 100); //Green
                color_wipe(neopixel_color_rgb(0, 127, 31), 100); //Turquoise
                color_wipe(neopixel_color_rgb(0, 0, 255), 100); //Blue
                color_wipe(neopixel_color_rgb(127, 0, 127), 100); //Purple
            }
        } while (!next_pattern);
        set_next_pattern_false(1);


#ifndef __AVR_ATtiny202__
        for (uint8_t i = 0; i < 8; i++) {
            if (!restoring_from_eeprom) {
                // Turn off all leds first
                for (uint8_t i = 0; i < neopixel_get_num_pixels(); i++)
                    neopixel_set_pixel_color_c(i, 0);
                neopixel_show();
                color_wipe_on_off(COLORS(i), 60);
            }
            set_next_pattern_false(2 + i);
        }
        for (uint8_t i = 0; i < 8; i++) {
            if (!restoring_from_eeprom) {
                // Turn off all leds first
                for (uint8_t i = 0; i < neopixel_get_num_pixels(); i++)
                    neopixel_set_pixel_color_c(i, 0);
                neopixel_show();
                color_wipe_on_off_spiral(COLORS(i), 60);
            }
            set_next_pattern_false(3 + 7 + i);
        }

        for (uint8_t i = 0; i < 8; i++) {
            if (!restoring_from_eeprom)
                color_set(COLORS(i));
            set_next_pattern_false(4 + 7 + 7 + i);
        }
#endif

        // Send a theater pixel chase in...
        if (!restoring_from_eeprom)
            theatre_chase(neopixel_color_rgb(127, 127, 127), 50); // White
        set_next_pattern_false(5 + 7 + 7 + 7 + 0);
        if (!restoring_from_eeprom)
            theatre_chase(neopixel_color_rgb(255, 0, 0), 50); // Red
        set_next_pattern_false(5 + 7 + 7 + 7 + 1);
        if (!restoring_from_eeprom)
            theatre_chase(neopixel_color_rgb(0, 255, 0), 50); // Green
        set_next_pattern_false(5 + 7 + 7 + 7 + 2);
        if (!restoring_from_eeprom)
            theatre_chase(neopixel_color_rgb(0, 0, 255), 50); // Blue
        set_next_pattern_false(5 + 7 + 7 + 7 + 3);

        if (!restoring_from_eeprom)
            rainbow(20);
        set_next_pattern_false(6 + 7 + 7 + 7 + 3);

#ifndef __AVR_ATtiny202__
        if (!restoring_from_eeprom)
            rainbow_cycle(40);
        set_next_pattern_false(7 + 7 + 7 + 7 + 3);
#endif

        if (!restoring_from_eeprom)
            rainbow_random_cycle(40);
        set_next_pattern_false(8 + 7 + 7 + 7 + 3);

        if (!restoring_from_eeprom)
            theatre_chase_rainbow(50);
        set_next_pattern_false(9 + 7 + 7 + 7 + 3);
    }
}
